/*-------------------------------------+-------------------------------------*/
/*                                 evaluate.c                                */

#include "board.h"

const unsigned long fac =    2416UL;
const unsigned long inc =  374441UL;
const unsigned long max = 1771875UL;

static signed long random (unsigned long limit)
{
   static unsigned long ran;

   ran = (ran * fac + inc) % max;
   return ran % (limit+limit) - limit;
}

/*-------------------------------------+-------------------------------------*/

long evaluate0 (void)
{
   return random (1000);
}

/*-------------------------------------+-------------------------------------*/

static long material_value[] = {1000, 3250, 3500, 5000, 9000, 0};

long evaluate1 (void)
{
   long result = 0;
   for (unsigned i=0; i<16; i++) {
      if (piecelist[WHITE][i].loc)
         result += material_value[piecelist[WHITE][i].type];
      if (piecelist[BLACK][i].loc)
         result -= material_value[piecelist[BLACK][i].type];
   }
   return result + random (10);
}

/*-------------------------------------+-------------------------------------*/

static inline long position (player col, unsigned n)
{
   return (piecelist[col][n].loc) ? piecelist[col][n].loc - board : 0;
}

long evaluate2 (void)
{
   long result = 0;
   for (unsigned i=0; i<16; i++) {
      if (piecelist[WHITE][i].loc)
         result += material_value[piecelist[WHITE][i].type];
      if (piecelist[BLACK][i].loc)
         result -= material_value[piecelist[BLACK][i].type];
   }

   // King backward
   result -= position (WHITE, 4) + position (BLACK, 4);
   // O-O prefered
   result += (piecelist[WHITE][4].loc==board+ 6) ? 500 : 0;
   result -= (piecelist[BLACK][4].loc==board+62) ? 500 : 0;

   result += position (WHITE, 11) + position (BLACK, 11); // d-pawn forward
   result += position (WHITE, 12) + position (BLACK, 12); // e-pawn forward
   result -= position (WHITE, 13) + position (BLACK, 13); // f-pawn backward
   result -= position (WHITE, 14) + position (BLACK, 14); // g-pawn backward
   result -= position (WHITE, 15) + position (BLACK, 15); // h-pawn backward

   // Knights and Bishops forward
   if (piecelist[WHITE][6].loc==board+ 6 || piecelist[WHITE][6].loc==board+23)
      result -= 100;
   if (piecelist[BLACK][6].loc==board+62 || piecelist[BLACK][6].loc==board+47)
      result += 100;
   result += position (WHITE, 2) + position (BLACK, 2); // Queen side Bishop
   result += position (WHITE, 5) + position (BLACK, 5); // King side Bishop

   // Silly, isn't it?

   return result + random (10);
}

/*-------------------------------------+-------------------------------------*/

//                                  Insert
//                                   YOUR
//                                 function
//                                   HERE
//                                    ...

/*-------------------------------------+-------------------------------------*/

struct {
   const char * name;
   long (*pointer)(void);
} efnlist[] = {
   { "Rand",  evaluate0 },
   { "Take",  evaluate1 },
   { "Eva",   evaluate2 },
   /*   ... and here.   */
   { NULL, NULL }
};

