/*-------------------------------------+-------------------------------------*/
/*                                 hardbord.c                                */

#include "board.h"

struct stack {
   stack * previous;
   char oldboard[64];
   state specials;
   unsigned move50;
   unsigned move_n;
   unsigned variation;
   piece * p0;
   field * f0;
   piece * p1;
   field * f1;
   object promotion;
};

struct move {
   piece * p0;
   field * f1;
   object promotion : 6;
   boolean check    : 1;
   boolean mate     : 1;
   move * next;
};


class hardboard {
   stack * sp;
   unsigned move50;
   unsigned move_n;
   unsigned variation;
   void push (piece *, field *, object = PAWN);
   void pop (void);

   const char * lang;
   void insert (unsigned, player, object);
   boolean setup (const char *);

   unsigned n_moves;
   boolean escape (piece *, field *, object = PAWN);
   boolean ismate (void);
   void trial (piece *, field *, object = PAWN);
   void make_movelist (void);

   char (* lanlist)[11];
   char (* sanlist) [8];
   boolean equal (const char *, const char *);
   unsigned lanhit (const char *, unsigned *);
   unsigned sanhit (const char *, unsigned *);
   void make_lanlist (void);
   void make_sanlist (void);
public:
   hardboard (const char * = "", const char * = "");
   void setlang (const char *);
   unsigned execute (const char *);
   void execute (piece *, field *, object = PAWN);
   const char * expand (piece *, field *, object = PAWN);
   void undo (void);
   void start_variation (void);
   void end_variation (void);
   move * movelist;
   boolean drawn_by_3 (void);
   boolean drawn_by_50 (void);

   void show_fen (void);
   void show_board (void);
   void show_lan (void);
   void show_san (void);
   ~hardboard (void);
};


static inline char present (piece * pp)
{
   return (pp) ? (pp->dark) ? char (-1-pp->type) : char (+1+pp->type) : '\0';
}

static inline stack * spush (stack * previous,
   unsigned move50, unsigned move_n, unsigned variation,
   piece * p0, field * f0, piece * p1, field * f1, object promotion)
{
   stack * result = new stack;
   result->previous = previous;
   for (unsigned i=0; i<64; i++)
      result->oldboard[i] = present (board[i].occ);
   result->specials = specials;
   result->move50 = move50;
   result->move_n = move_n;
   result->variation = variation;
   result->p0 = p0;
   result->f0 = f0;
   result->p1 = p1;
   result->f1 = f1;
   result->promotion = promotion;
   return result;
}

static inline stack * spop (stack * sp)
{
   stack * result = sp->previous;
   delete sp;
   return result;
}

static inline void movepiece (field * f0, field * f1)
{
   f1->occ = f0->occ;
   f1->occ->loc = f1;
   f0->occ = NULL;
}

void hardboard::push (piece * p0, field * f1, object promotion)
{
   if (f1==specials.epfield && p0->type==PAWN)
      movepiece (f1->pnext[!color][0], f1);

   field * f0 = p0->loc;
   piece * p1 = f1->occ;

   if (p0->type == KING) {
      if (f0==board+ 4) {
         if (f1 == board+ 2) movepiece (board+ 0, board+ 3);
         if (f1 == board+ 6) movepiece (board+ 7, board+ 5);
      }
      if (f0==board+60) {
         if (f1 == board+58) movepiece (board+56, board+59);
         if (f1 == board+62) movepiece (board+63, board+61);
      }
   }

   sp = spush (sp, move50, move_n, variation,
      p0, p0->loc, f1->occ, f1, promotion);
   if (specials.OOwhite  && (f0==board+ 4 || f0==board+ 7 || f1==board+ 7))
      specials.OOwhite  = FALSE;
   if (specials.OOOwhite && (f0==board+ 4 || f0==board+ 0 || f1==board+ 0))
      specials.OOOwhite = FALSE;
   if (specials.OOblack  && (f0==board+60 || f0==board+63 || f1==board+63))
      specials.OOblack  = FALSE;
   if (specials.OOOblack && (f0==board+60 || f0==board+56 || f1==board+56))
      specials.OOOblack = FALSE;
   if (p0->type)
      specials.epfield = NULL;
   else
      if (color==WHITE && f0<board+16 && f1>=board+24 ||
         color==BLACK && f0>=board+48 && f1<board+40)
         specials.epfield = f0->pnext[color][0];
      else
         specials.epfield = NULL;
   if (p0->type==PAWN || f1->occ)
      move50 = 0;
   else
      move50++;
   if (color)
      move_n++;

   f0->occ = NULL;
   f1->occ = p0;
   p0->loc = f1;
   if (p1)
      p1->loc = NULL;
   if (promotion)
      p0->type = promotion;

   color = (color) ? WHITE : BLACK;
   return;
}

void hardboard::pop (void)
{
   color = (color) ? WHITE : BLACK;

   sp->f0->occ = sp->p0;
   sp->f1->occ = sp->p1;
   sp->p0->loc = sp->f0;
   if (sp->p1)
      sp->p1->loc = sp->f1;
   if (sp->promotion)
      sp->p0->type = PAWN;

   specials = sp->specials;
   move50 = sp->move50;
   move_n = sp->move_n;
   variation = sp->variation;

   if (sp->p0->type == KING) {
      if (sp->f0==board+ 4) {
         if (sp->f1 == board+ 2) movepiece (board+ 3, board+ 0);
         if (sp->f1 == board+ 6) movepiece (board+ 5, board+ 7);
      }
      if (sp->f0==board+60) {
         if (sp->f1 == board+58) movepiece (board+59, board+56);
         if (sp->f1 == board+62) movepiece (board+61, board+63);
      }
   }

   if (sp->f1==specials.epfield && sp->p0->type==PAWN)
      movepiece (sp->f1, sp->f1->pnext[!color][0]);

   sp = spop (sp);
   return;
}

void hardboard::insert (unsigned i, player col, object type)
{
   if (i<64 && ! board[i].occ) {
      unsigned j;
      for (j=0; j<8; j++)
         if (piecelist[col][j].type==type && piecelist[col][j].loc==NULL) {
            piecelist[col][j].loc = board+i;
            board[i].occ = piecelist[col]+j;
            break;
         }
      if (j==8)
         for (j=8; j<16; j++)
            if (type!=KING && piecelist[col][j].loc == NULL) {
               piecelist[col][j].type = type;
               piecelist[col][j].loc = board+i;
               board[i].occ = piecelist[col]+j;
               break;
            }
   }
   return;
}

static inline char low (char c)
{
   return ('A'<=c && c<='Z') ? c+'a'-'A' : c;
}

boolean hardboard::setup (const char * s)
{
   unsigned i;

   for (i=0; i<64; i++)
      board[i].occ = NULL;
   for (i=0; i<16; i++)
      piecelist[WHITE][i].loc = piecelist[BLACK][i].loc = NULL;
   for (i=8; i<16; i++)
      piecelist[WHITE][i].type = piecelist[BLACK][i].type = PAWN;

   for (i=56; *s && *s!=' '; s++)
      if      (*s == lang[PAWN])   insert (i++, WHITE, PAWN);
      else if (*s == lang[KNIGHT]) insert (i++, WHITE, KNIGHT);
      else if (*s == lang[BISHOP]) insert (i++, WHITE, BISHOP);
      else if (*s == lang[ROOK])   insert (i++, WHITE, ROOK);
      else if (*s == lang[QUEEN])  insert (i++, WHITE, QUEEN);
      else if (*s == lang[KING])   insert (i++, WHITE, KING);
      else if (*s == low(lang[PAWN]))   insert (i++, BLACK, PAWN);
      else if (*s == low(lang[KNIGHT])) insert (i++, BLACK, KNIGHT);
      else if (*s == low(lang[BISHOP])) insert (i++, BLACK, BISHOP);
      else if (*s == low(lang[ROOK]))   insert (i++, BLACK, ROOK);
      else if (*s == low(lang[QUEEN]))  insert (i++, BLACK, QUEEN);
      else if (*s == low(lang[KING]))   insert (i++, BLACK, KING);
      else if (*s == '/')          i -= 16;
      else if ('1'<=*s && *s<='8') i += *s-'0';
      else ;  /* Unknown char in FEN */
   if (*s) s++;
   if (! piecelist[WHITE][4].loc || ! piecelist[BLACK][4].loc)
      return FALSE;

   if (! *s || *s=='W')
      color = WHITE;
   else
      color = BLACK;
   if (*s) s++;
   if (*s) s++;

   specials.OOwhite  = FALSE;
   specials.OOOwhite = FALSE;
   specials.OOblack  = FALSE;
   specials.OOOblack = FALSE;
   if (s[0]=='n' && s[1]=='i' && s[2]=='l')
      s += 3;
   else
      for ( ; *s && *s!=' '; s++)
         if      (*s == lang[KING])       specials.OOwhite  = TRUE;
         else if (*s == lang[QUEEN])      specials.OOOwhite = TRUE;
         else if (*s == low(lang[KING]))  specials.OOblack  = TRUE;
         else if (*s == low(lang[QUEEN])) specials.OOOblack = TRUE;
         else ;  /* Unknown rochade */
   if (*s) s++;
   if (piecelist[WHITE][4].loc != board+ 4)
      specials.OOwhite = specials.OOOwhite = FALSE;
   if (piecelist[BLACK][4].loc != board+60)
      specials.OOblack = specials.OOOblack = FALSE;
   if (! board[ 0].occ || ! board[ 0].occ->dark==WHITE ||
      ! board[ 0].occ->type==ROOK)
      specials.OOOwhite = FALSE;
   if (! board[ 7].occ || ! board[ 7].occ->dark==WHITE ||
      ! board[ 7].occ->type==ROOK)
      specials.OOwhite  = FALSE;
   if (! board[56].occ || ! board[56].occ->dark==BLACK ||
      ! board[56].occ->type==ROOK)
      specials.OOOblack = FALSE;
   if (! board[63].occ || ! board[63].occ->dark==BLACK ||
      ! board[63].occ->type==ROOK)
      specials.OOblack  = FALSE;

   specials.epfield = NULL;
   if (s[0]=='n' && s[1]=='i' && s[2]=='l')
      s += 3;
   else if ('a'<=s[0] && s[0]<='h' && (s[1]=='3' || s[1]=='6')) {
      specials.epfield = board + (s[0]-'a') + (s[1]-'1') * 8;
      s += 2;
      if (specials.epfield->occ ||
         ! specials.epfield->pnext[!color][0]->occ ||
         ! specials.epfield->pnext[!color][0]->occ->dark != color ||
         ! specials.epfield->pnext[!color][0]->occ->type == PAWN)
         specials.epfield = NULL;
   }
   if (*s) s++;

   for (move50 = 0; *s && *s!=' '; s++)
      move50 = 10 * move50 + *s - '0';
   if (*s) s++;
   for (move_n = 0; *s && *s!=' '; s++)
      move_n = 10 * move_n + *s - '0';
   if (*s) s++;
   if (! move_n)
      move_n = 1;

   color = (color) ? WHITE : BLACK;
   boolean result = (attacked (piecelist[color][4].loc)) ? FALSE : TRUE;
   color = (color) ? WHITE : BLACK;
   return result;
}

hardboard::hardboard (const char * fen, const char * language)
{
   lang = (*language) ? language : "PNBRQK";
   if (! *fen)
      fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR W KQkq nil 0 1";
   if (! setup (fen)) {
      lang = "PNBRQK";
      fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR W KQkq nil 0 1";
      setup (fen);
   }

   sp = NULL;
   variation = 0;
   movelist = NULL;
   lanlist = NULL;
   sanlist = NULL;
   make_movelist ();
   make_lanlist ();
   make_sanlist ();
   return;
}

void hardboard::setlang (const char * language)
{
   lang = language;
   make_lanlist ();
   make_sanlist ();
   return;
}

unsigned hardboard::execute (const char * s)
{
   unsigned result, movenum;

   if ((result=sanhit(s,&movenum)) == 1 || (result=lanhit(s,&movenum)) == 1) {
      move * m = movelist;
      while (movenum--)
         m = m->next;
      push (m->p0, m->f1, m->promotion);
      make_movelist ();
      make_lanlist ();
      make_sanlist ();
   }
   return result;
}

void hardboard::execute (piece * p0, field * f1, object promotion)
{
   for (move * m=movelist; m; m=m->next)
      if (m->p0==p0 && m->f1==f1 && m->promotion==promotion)
         push (m->p0, m->f1, m->promotion);
   make_movelist ();
   make_lanlist ();
   make_sanlist ();
   return;
}

const char * hardboard::expand (piece * p0, field * f1, object promotion)
{
   unsigned n = 0;
   for (move * m=movelist; m; m=m->next, n++)
      if (m->p0==p0 && m->f1==f1 && m->promotion==promotion)
         return sanlist[n];
   return NULL;
}

void hardboard::undo (void)
{
   if (sp && sp->variation==variation) {
      pop ();
      make_movelist ();
      make_lanlist ();
      make_sanlist ();
   }
   return;
}

void hardboard::start_variation (void)
{
   variation++;
   return;
}

void hardboard::end_variation (void)
{
   if (variation) {
      while (sp && sp->variation==variation)
         pop ();
      variation--;
      make_movelist ();
      make_lanlist ();
      make_sanlist ();
   }
   return;
}

static unsigned deja_vu (stack * s)
{
   if (specials.OOwhite != s->specials.OOwhite  ||
      specials.OOOwhite != s->specials.OOOwhite ||
      specials.OOblack  != s->specials.OOblack  ||
      specials.OOOblack != s->specials.OOOblack ||
      specials.epfield  != s->specials.epfield)
      return 0;
   for (unsigned i=0; i<64; i++)
      if (s->oldboard[i] != present(board[i].occ))
         return 0;
   return 1;
}

boolean hardboard::drawn_by_3 (void)
{
   unsigned repetitions = 1;
   unsigned m50 = move50;
   stack * s = sp;

   while (TRUE) {
      if (! m50 || ! s)
         break;
      m50 = s->move50;
      s = s->previous;
      if (! m50 || ! s)
         break;
      repetitions += deja_vu (s);
      m50 = s->move50;
      s = s->previous;
   }
   return (repetitions < 3) ? FALSE : TRUE;
}

boolean hardboard::drawn_by_50 (void)
{
   return (move50 < 100) ? FALSE : TRUE;
}

hardboard::~hardboard (void)
{
   for (stack * helps=sp; helps; helps=sp) {
      sp = sp->previous;
      delete helps;
   }
   for (move * helpm=movelist; helpm; helpm=movelist) {
      movelist = movelist->next;
      delete helpm;
   }
   delete lanlist;
   delete sanlist;
   return;
}

